/*********************************************************\
*  AVR MegaSquirt - 2003                                  *
*  (C) 2003 [Michael Kristensen, mik@caffrey.dk]          *
***********************************************************
*  -- History --                                          *
*    release 0.1 - 06-03-2003                             *
*            http://caffrey.dk/megasquirt                 *
*    initial Motorola version by Bowling & Grippo (2002)  *
*            http://www.bgsoflex.com/megasquirt.html      *
\*********************************************************/

// Type declarations global to everything

#ifndef GLOBAL_H
#define GLOBAL_H

#pragma GCC diagnostic ignored "-Wdeprecated-declarations"
#pragma GCC diagnostic ignored "-Wignored-qualifiers"

#include <inttypes.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
// EXTERN will diminish in vems.c and used as extern in other files
#ifndef EXTERN
#define EXTERN extern
#endif

typedef uint16_t psize_t;

/* special hardware is defined in my_make (see doc/)
*/
#define KNOCK_8101  // knock: yes, please
#define MCP3208

#ifndef IGNCHANNELS
#define IGNCHANNELS 10
#endif

#ifndef BOOSTCONTROL_GEARS
// default is to compile with BOOSTCONTROL unless BOOSTCONTROL_GEARS=0 is in my_make
// actually "6-gear" boostcontrol:
#define BOOSTCONTROL_GEARS 6
#endif

#if(BOOSTCONTROL_GEARS > 0)
#define BOOSTCONTROL // convenience define, use this to check if BOOSTCONTROL is needed
#endif

#define HWMAP
#include "hwdef_v3.h"

#ifndef true
#define true 1
#endif

#ifndef false
#define false 0
#endif

/***********************************************************************/
//flags
// #define tenth_second 0
#define second       1

// the high resolution now and now_upper (16+16 bits) is in eventqueue
struct time_t {
// these are clocked from handler2k:
//  uint8_t iac_clk;
//  uint8_t tsec_clk; // 100 ms clock
//  uint16_t h2k;   // h2khandler counter (800 usec) is in benchstats. Move here if you need it.
// these are clocked from userspace:
  uint8_t tsec;  // seconds /10
  uint16_t sec;  // seconds
#ifndef WBO2
	uint8_t adc_clk;
#endif
};

// rtc is not volatile any longer. Some fields are only set and beeing watched
// from userspace (like flags and most); adc_clk only from handler2k()
extern struct time_t rtc;
/***********************************************************************/

#define enabled   0 // squirt
#define scheduled 1 // sched
#define firing    2 // firing

// toothlen.c defines: review toothlen_simple() before changing
#define TOOTHREL_MISSING (0)
#define TOOTHREL_SINGLE (1)
#define TOOTHREL_ERROR (0xFF)

struct squirt_t {
	uint16_t pwcalc;    // pulsewidth in 4 usec resolution
	uint16_t stopat;	// when to stop (4 usec time)
};

// volatile dropped, interrupt will reload the values anyway
#define INJECTORS 8

// inj_stage2_start_tps = FF (100%) means only enable above 100% TPS => disable
#define INJECTOR_STAGING_ENABLED  (config.inj_stage2_start_tps != 0xFF)

#define PRIMARY_INJ_LAST (config.alternate & 7)
// bad implementation required that SECONDARY_INJ_CHANNELS_NUM == PRIMARY_INJ_LAST
// That could not do any good, just break things
// #define SECONDARY_INJ_CHANNELS_NUM  ((config_ext.inj_stage2_conf >> 1) & 0x07)
// see injector_handler() for proper implementation that also works for 5-6 primary injectors

#ifdef SOFTRPM
#ifdef NOIGN
#error "Soft RPM limiters need to use ignition, please undefine NOIGN"
#endif
#define IGNCUTLVL_MAX 16
#endif


extern struct squirt_t inj_port[INJECTORS];

/***********************************************************************/

//status
#define running 0  // engine running. Should NOT be set upon first noisepulse, just when trigger looks sane => than only inject/spark when this is set
#define crank   1  // cranking
#define startw  2  // afterstart enrichment. Since we use etime.asecount anyway, it would be smarter to use ! _BV(crank) instead
#define warninglight  3  // was warmup earlier
#define tpsaen  4  // tps acceleration
#define tpsden  5  // tps deceleration
#define wbo2heateron  6  // wbo2heater nonzero duty
#define idle    7  // idle mode

//status_ext
//#define new_rpm 0  // was flag "rpm recalc needed"
#define baro_problem     1  // somehow the barometric pressure is way off
#define o2_not_ready     2  // this is set for the first 30 seconds
//UNUSED #define ...   3
#define crank_enable     4  // allowed to enter crank mode
#define als_active	     5  // antilagsystem is active
#define als_switch_activated  6 // true if ALS switch is activated
#define individual_allowed 7 // MegaTune flags to allow individual power recalc

#define is_als_active() (engine.status_ext & _BV( als_active ))

//status_trig cleared in engine_off(): engine.status_trig = 0;
// #define engine_in_sync   0	// now using cam_sync bit instead
// #define wheel_in_sync    1  // check etime.tooth_cnt
#define PRIM_HIGH_AT_SECTRIG 0 //sample primary input state when sectrig received
#define n2o_active	1
//#define wheel_error      2  // error during wheel decoding
#define exhaust_pulse_seen 2 //exhaust rising edge seen from last camsync pulse
#define cam_syncsticky   3  // for timing.c cam_trigger_sync() or instead of engine_in_sync. Only cleared in fuelpump_off
#define cam_sync2maserati	 4	// sticky signal back from schedule_dwell_event() to MASERATI
#define ign_evnt_drp	 5  // signal that one event has been dropped
#define cam_sync3	 6	// for multitooth.c
#define injector_staging_active   7 // needed for hysteresis activation

#define cam_sync_mask  ( _BV(cam_syncsticky) | _BV(cam_sync3) )
#define cam_sync_maskcheat  ( _BV(cam_sync3) )

// status4 bits (only set from userspace, no need for cli/sei):
#define launchinput_active 0	// the input button, not the actual retard / sparkcut function
#define launchfunction_active    1 // launch control FUNCTION (retard and/or sparcut) is active
#define STATUS4_CONFSW 2//config sw status 0= config A; 1= config B

#define prime_was		7 // used for flag if prime pulse was

#ifdef N2O_RETARD
//
// bits for config_ext.n2o_settings
//
#define n2o_retard_enable       0
#define launch_sw_deactivate    1
#define n2o_shiftcut_disables_n2o 2

//
// macros for n2o
//
#define n2o_retard_disabled() \
    ((config_ext.n2o_settings & _BV(n2o_retard_enable)) == 0)

#define n2o_launch_sw_deactivation_disabled() \
    ((config_ext.n2o_settings & _BV(launch_sw_deactivate)) == 0)

#define shiftcut_disables_n2o() (config_ext.n2o_settings & _BV(n2o_shiftcut_disables_n2o))

#define during_actual_shiftcut() (engine.shiftcut_state & ( _BV( IGNCUT_SHIFTCUT ) | _BV( FUELCUT_SHIFTCUT ) ) )

// input handling: reuse launchswitch handler
#define n2o_input_active() (input_switch_deactivated(config_ext.n2o_input & 0xF) == INPUTSWITCH_ACTIVE)

#define launch_sw_disables_n2o() \
      ((n2o_launch_sw_deactivation_disabled() == 0) && is_launchfunction_active())

#endif

#ifndef OVERRUN_DELAY
// do NOT set this higher than 0. It kills sane behaviour with several engines.
// define in your my_make, or make it adjustable from vemstune if needed.
#define OVERRUN_DELAY	0
#endif
// strig_validation bits
//bit 0,1 camshaft measure count for Subaru and Suzuki VVTI
#define CAMSHAFT_MEASURE_PULSE_CNT_MASK ( _BV(0) + _BV(1) )
#define CAM_CHECK_POLARITY 0
#define NISSAN_DIVIDER 0
#define NISSAN_TRIGGER 2
#define EXTRA_PULSE_TRIGGER 3
#define SUZUKI_TRIGGER      4
#define ls_1_forgiving   5       // if 1, LS-1 48+1 forgiving tooth_cnt == 46 when primary input high at sectrig
#define STRIG_VALIDATION_SHUTDOWN 6
#define STRIG_VALIDATION_FILTER 7

//status_output
#define STATUS_OUT_MISC1_BIT 0
#define STATUS_OUT_MISC2_BIT 1
// 2..5 unused
#define STATUS_OUT_WOT_BIT   6
#define STATUS_OUT_FUEL_BIT  7

struct engine_t {
  uint8_t prohibit_status; //prohibit ignition and injection if version not correct or ibutton not found
  uint8_t status;      // bit defines above
  uint8_t status_ext;  // bit defines above! Not changed from interrupt (cli-sei not needed, but won't hurt).
  uint8_t status_trig; // trigger status. Also changed from interrupt so userspace MUST cli-sei.
  uint8_t trigger_error; // reason of error 1: too many ptrig 2: not enough ptrig 3: too many sectrig 4: not enough sectrig 5: sec pos bad
  uint8_t batt;		// battery voltage (scaled as a fraction of 30V. 128=15V)
  uint16_t kpa250;		// MAP value in units of 250Pa (note: the "k" in kpa is historical)
  uint8_t coolant;     // Coolant temperature in Degrees C plus 100
  uint8_t mat;         // manifold air temperature -100..+155C
  uint16_t egt;		// Exhaust gas temperature [ Celsius ]
  uint16_t egt2;		// Exhaust gas2 temperature [ Celsius ]
  uint8_t tps;         // throttle posision (0..255)
  uint8_t last_tok;    // same as tps, or MAP value in units of KPa, last reading
  uint16_t rpm_period16;	// rpm period lower 16 bit (in 4 usec resolution). See rpm_period_upmost and ENGINE_RPM_PERIOD24. Use cli-sei from userspace to fetch atomically
  uint8_t rpm_period_upmost; // upmost byte of rpm_period16 (normally 0..3). FF means invalid. FE means first pulse already seen.
  uint8_t rpm;         // computed engine rpm/100 (stupid origins in motorola megasquirt, comm compatibility likes this)
  uint8_t kpa_or_tps; // tps below speeddensity_minkpa threshold, kpa above it
  uint8_t rpm_new_rpm; // 0 means RPM recalc needed. Wasting 7 bits for speed and clarity. See flag_new_rpm()
  uint16_t rpm_vhr;		// computed very high res RPM (real RPM, like 8000)
  uint16_t calculated_fuelpw; // [4 usec] fuelpw calculated even during cranking, when inj_port is all 0;
  uint16_t sparkat;	// absolute time when (appr) spark will come. Set at ign_act(). [4usuc]
  uint8_t acc_retard; // weighted retard for acceleration enrichment
	uint8_t fuelcut_learnt; // remember partial fuelcut to apply when TPS pressed
	uint8_t tok_diff; // tps or kpa dot (change rate, dV/dt)
  uint8_t boost_tps_downsc;// boost tps downscaling value
  uint8_t target_last_refresh;//boost target refresh timestamp
  uint8_t status_output; // ONLY changed from userspace. output states: 7:fuelcontrol, 6WOT actuator, 5 fuelpump 1:Misc2, 0:Misc1
  uint8_t misc_countup; // countup every 100 msec (until reaching 255, also initial value). misc output effected. Currently from launch-inactivation, later configurable
	uint8_t cltfan_dontchangecountdown; // coolant fan output must not change if nonzero (and cuntdown every 100msec)
#ifdef ALS
  uint8_t als_countdown; // countdown every 100 msec. ALS disabled when reaches 0
#endif
#ifndef NOIGN
  uint16_t ignadv;	// ignition advance + IGN_BALANCE (0.25 crankdeg)
//  uint32_t cam_period; // cam period (in 4 usec resolution)
	uint8_t ignumsc; // http://tiki.vems.hu/tiki/tiki-index.php?page=IgnitionTiming
	uint8_t ignumdc;
	uint16_t sameoutput_mintooth_scaler;//use it to limit the dwell
	uint16_t dwell;		// in 4 usec resolution. this is now limited
	uint16_t trig2spark;	// time from trigger to spark [4 usec] (NOT rpm_period * trig2spark / 4096 !)
	uint16_t proposed_trig2spark; // 4 usec (not angle "factor" like earlier)
	uint8_t spark_phase;	// actually toothcnt relative to the reference tooth for given cyl
	uint8_t proposed_spark_phase;
	uint8_t dwell_phase;
	uint8_t proposed_dwell_phase;
	uint16_t trig2dwell;	// time from trigger to dwell [4 usec]
	uint16_t proposed_trig2dwell;  // 4 usec
	uint16_t rotary_trailing_spark_delay; //trailing spark delay in rotary mode, 4 usec
	uint16_t pwm_delay;	// delay (4 usec) before going to pwm-mode (injpwmt) SHOULD depend on Vbatt
	// uint8_t cylinders;	// use engine_cylinders
	uint8_t status4;	//see defines below, bit0: launchinput_active; bit1: launchfunction active; bit2: config sw state
	uint8_t launch_rpm_trim; // added to launch rpm limit by anytrim [100rpm]
#ifdef SOFTRPM
  uint8_t randomseed;	// random seed for soft rpm limiter
  uint8_t sparkcutlvl;  // spark cut level 0..16
#endif
#endif
  uint8_t injector_bitmask; // bitmask used when firing all injectors simultaneously
  uint8_t shiftcut_state;	// only changed from userspace;
  // lower 2 bit is state; bit7:2 are flags. See below.

#ifdef BOOSTCONTROL
	uint16_t boost_stopat;	// time of boostcontrol actuation event
	uint16_t boostalt_stopat; // time of boostcontrol ALT actuation event
	uint16_t boost_ref_dc;       // (0 .. 255) << 3 means 0 to 100%
 // linear interpolated (according to RPM), but not downscaled according to tps:
	uint8_t boostmax_target; // 2kPa
	uint8_t boost_command_duty; // 0 .. 255 means 0 to 100%
	uint8_t boostalt_command_duty; // 0 .. 255 means 0 to 100%
	// uint8_t anytrim1; // if we want to save it after lookup and interpolation. 0..255 meaning depends on strategy
#endif
	uint8_t ptrig_count;	// incremented by PRIMARY_TRIGGER handler in the subaru branch.
	uint8_t strig_count;	// incremented by SECONDARY_TRIGGER in the subaru branch only. Auditrigger decrements in primary handler.
	uint8_t strig_count_old; // stores the old value of strig_count for Nissan trigger
	uint8_t syncaid_threshold; // only used for AUDITRIGGER at the moment
	uint8_t individual_power; // filled by flag_rpm();
	uint8_t partialfuelcut;  // partial fuelcut 0..15 (16 is no fuel at all, but currently not used)
	uint8_t partialfuel_enrich; //at soft fuel cut we can enrich fuel too for partial fuel
	uint16_t softcut_time;  //[8ms]
#ifndef  PS2KEYBOARD
	uint8_t revlimit;
	uint8_t gear; // 0: gear1 .. 5: gear6
#ifdef SECOND_VSS_EXPERIMENTAL
	uint8_t launch_gear;
	uint8_t launch_counter;
#endif // SECOND_VSS_EXPERIMENTAL
#endif
	// time-based launch will be possible even when PS2 CLOCK is used for wheelspeed
	uint8_t launchinact_time; // saved timepoint of launchswitch inactivation, 64msec
#ifdef KNOCK_ALTERNATIVE
	uint8_t knock_noise_thres;// value from knock noise table
#endif
	uint8_t injstate_bitmap; /** each bit is the state of an injector channel. bit = 1 injector active
// * next INJSTOP event should turn off the injector if injector.stopat elapsed
// * Any INJSTART can be neglected if injector is active already
// * Note; we always execute PWM-ing on staring injector
// * Note2: use BV_rev(a) to get a mask for channel a */
	uint8_t injnum; // injector bank(group) number(index) to fire
#ifdef INJ_ANGLE
	uint8_t inj_stopangle;
	uint16_t trig2injstop; // [4 usec] delay from triggertooth to inj stop. Might be neglected if pw>trig2injstop
//	uint8_t inj_fire_mask; //banks wich we need to fire now (to prevent event skipping)
#endif

#ifdef TOYOTA_1JZ_VVTI
	uint8_t vvt_target; //final target camshaft position
	uint8_t vvt_position; // camshaft position
	uint8_t vvt_map; // camshaft angle target, map value
	uint8_t vvt_tooth; // 1: tooth now.
#endif
//  uint8_t alien_ignadv;  // measured ignition advance (toothwheel + coil trigger)
//  uint8_t alien_dwell;  // measured ignition dwell

#ifdef FUEL_CONS_CALCS
//	uint16_t l_ph;
//	uint32_t l_ph_avg;
//	uint16_t speed_avg;
	uint16_t l_100km_avg;   // liter / 10000 km average consumption SINCE boot
//	uint16_t pw_fuelonly;

	uint32_t distance;			//distance in millimeters, 4200km is enough
	uint32_t fuelcons;			//fuel in 0.00001cc, 42l is enough
#endif

#ifdef TOYOTA_CUSTOM
	uint8_t clt_dc;
	uint8_t canister_dc;
#endif

#ifdef CHARGE_TEMP
	uint8_t charge_temp; //charge temp [0.5Deg] -30 .. 97.5 DegC
#endif
#ifndef MAINLOOP_TASKER //not needed for tasker, tasker uses mainlppwm.pwm_states & _BV(ML_PWM_TACH)
	uint8_t tachout_state; // effectively only bit 7 is used (easy to free 7 bits)
#endif
  uint16_t sectrig_posI;// time from sectrig to next toothcnt==0 [4us]
  uint32_t ptrig_captured_at0; // primtrig capture at tooth_cnt==0
//  uint16_t strig_diff; //difference between two sectrig, used for subaru angle control
//  uint32_t strig_capturedtime_prevall;// strig captured time in subaru branch
  #ifdef CAMSHAFT_ANGLE
#ifdef CAMSHAFT_ANGLE_EXHAUST
  uint16_t sectrig_posE;// time from exhtrig to next toothcnt==0 [4us]
  uint16_t cam_stopatE;//cam output stop at
  uint8_t cam_targetE;// sectrig target
  uint8_t extrig_angle;// negativ angle [2.8 deg]
  uint16_t camshaft_dcE; // camshaft actuator duty cycle

#endif
  uint16_t camshaft_dcI; // camshaft actuator duty cycle
  uint8_t sectrig_angleI;// negativ angle [90 / 32 = 2.8 deg]
  uint8_t cam_targetI;// sectrig target
  uint16_t cam_stopatI;//cam output stop at
#endif
#ifdef CAM_POS_CHECK
  uint32_t strig_next_mintime; //next valid secondary trigger minimum time[4us]
  uint32_t strig_next_maxtime; //next valid secondary trigger maximum time[4us]
#endif
#ifdef NERNST_ADC_FROM_USERSPACE
  uint8_t adc_userspace_flag; // 1=>userspace responsible; 0=>interrupt responsible
#endif

#if IGNCHANNELS>8
uint8_t ignum_dwellarr[IGNCHANNELS];	// 1 byte for each bit
#else
uint8_t ignum_dwell;	// flags if cyl is under dwell: 1 bit for each
#endif
};

EXTERN volatile struct engine_t engine;

// request RPM recalculation and injection:
#define flag_new_rpm() engine.rpm_new_rpm=0
static inline uint8_t userspace_TaS_new_rpm(void){ // "Test and Set": return true if RPM recalc required (also setting rpm_new_rpm)
    uint8_t ret=0;
    cli(); // not really necessary: wont help if we haven't consumed, and interrupt overwritten 0 with 0 before this point
    if (engine.rpm_new_rpm==0){
	engine.rpm_new_rpm=0xa7;
	ret=1;
    }
    sei();
    return ret;
}

// Use cli-sei from userspace to fetch atomically
#define ENGINE_RPM_PERIOD24 ((((uint32_t)engine.rpm_period_upmost)<<16) + engine.rpm_period16)
#define invalidate_rpm_period() engine.rpm_period_upmost = 0xFE
#define invalidate_rpm_periodFF() engine.rpm_period_upmost = 0xFF
// engine.status & _BV(running) could be better and cheaper condition instead
// upmost byte is 0:
#define isvalid_rpm_period() ( engine.rpm_period_upmost < 16)
#define is_rpm_periodFF() ( engine.rpm_period_upmost == 0xFF)

#ifdef  PS2KEYBOARD
#define get_wheelspeed() 0
#else
#define get_wheelspeed() (wheel_speed1.wspeed)
#endif

// engine.shiftcut_state bit1:0 state definitions
// state transitions: 0 => 1 => 3 => 2
// (don't change these unless you review all usage of shiftcut_state)
#define SHIFTCUT_STATE_NEGLECTNOISE 0
#define SHIFTCUT_STATE_INACTIVE 1
#define SHIFTCUT_STATE_ACTIVE 2
#define SHIFTCUT_STATE_ARMED 3

// engine.shiftcut_state bit7:2 bits:
#define FUELCUT_SHIFTCUT 7
// if any of these bits are set (by independent sources), we need 100% full fuelcut:
#define FUELCUT_MASK 0x80

// when config.overrun_fuelcut - config.overrun_fuelresume ) == 1000 or 1100 RPM difference
// we set engine.shiftcut_state & _BV(OVERRUN_PARTIALFUEL)
#define OVERRUN_PARTIALFUEL 6
#define smoothoverrun_configured() (((config.overrun_fuelcut - config.overrun_fuelresume) & 0xFE) == 10)

#define IGNCUT_SHIFTCUT 2
#define IGNCUT_REVCUT 3
// if any of these bits are set (by independent sources), we need igncut:
#define IGNCUT_MASK 0x0c

// I know these should be elsewhere ... feel free to cleanup
static inline void updateDwellPhaseFromProposal(void)
{
	engine.dwell_phase = engine.proposed_dwell_phase;
	engine.trig2dwell = engine.proposed_trig2dwell;
}

static inline void updateSparkPhaseFromProposal(void)
{
	engine.spark_phase = engine.proposed_spark_phase;
	// engine.trig2spark is not updated at low RPM, seen for 10+1 and cosworth
	// (these have long 72 and 90 degree teethgaps)
	// likely dispatcher.c disp_igndeact() issue
	engine.trig2spark = engine.proposed_trig2spark;
}

#define INVALID_TRIGGER_TOOTH 0x80
    /**
     * @return index of next coil to be fired (ignum upper bits)
     */
static inline uint8_t get_ignumsc(void){
        return  engine.ignumsc;
}

static inline uint8_t get_ignumdc(void){
        return engine.ignumdc;
}

static inline void set_ignumsc(uint8_t newval){
        engine.ignumsc = newval;
}

static inline void set_ignumdc(uint8_t newval){
	engine.ignumdc = newval;
}

// reset both set_ignumdc() and set_ignumsc()
static inline void reset_ignum(uint8_t newval){
	// TODO: should we check if isCoilActive() ?
	engine.ignumsc = engine.ignumdc = newval;
}

#if IGNCHANNELS>8
static inline uint8_t isCoilActive(void)
{
	uint8_t i;

	for( i=IGNCHANNELS; i;  ){
		if (engine.ignum_dwellarr[--i]  != 0) return 1; // dwell was started
	}
	// none started:
        return 0; // (get_ignumsc() == get_ignumdc()) ? 0 : 1;
}
#else
static inline uint8_t isCoilActive(void)
{
        return engine.ignum_dwell; // (get_ignumsc() == get_ignumdc()) ? 0 : 1;
}
#endif

// if( ! isCoilActive() ) ... engine.ignum = 0;
/* static inline void reset_ignum_if_inactive( void )
*{
*	uint8_t ei, t8;
*
*	ei = engine.ignum;
*	t8 = ei >> 4;
*	ei &= 0x0F;
*	ei ^= t8; // check with XOR if upper and lower nibble matches
*	if( ei == 0 ){ // they matched => isCoilActive() == false
*		engine.ignum = 0;
*	}
} */

#define AUDITRIGGER_NEVER_THISHIGH 0x8F
static inline uint8_t add_modulo(uint8_t a, uint8_t b, uint8_t total_tooth)
{
        uint8_t ret = a + b;
		/*
        if( ret >= AUDITRIGGER_NEVER_THISHIGH ){
		// >143: negative because of underflow;
		// note that up to 135 (or even 142) is perfect with AUDITRIGGER
		ret += total_tooth;
		return ret;
        }*/


        if( total_tooth ){
		while( ret >= total_tooth ){
        		ret -= total_tooth;
		}
        }
        return ret;
}

    /**
     * True, if toothcnt is same as searchedtooth, or just follows, like
     * toothcnt == searchedtooth +1
	* examples:
     * searchedtooth toothcnt
     * 0    1
     * 1    2
     * ...
     * 12   13
     * 13   15 YES, must be allowed, and possible
     * 15   16
     * 28    0  toothcount overflow, allowed
	* 57	 60 when total_tooth==120
	* 57	0
     * 14   15 Special, missing tooth14 should have been avoided !
     * @param allowedtooth
     * @param toothcnt
     * @return true if allowedtooth == toothcnt

	TODO: WHY THE FUCK WOULD WE NEED to allow toothcnt one later than searchedtooth?
	From multitooth.c, a condition for exact match is better

static inline uint8_t appr_same(uint8_t searchedtooth, uint8_t toothcnt, uint8_t total_tooth)
{
	int8_t allowedtooth = searchedtooth - toothcnt;
        if( allowedtooth <= 0 ){ // nice, there it is
            if( allowedtooth >= -3 ){ // -2 for N-1; -3 for 60-2
                return 1;
            }
        }

        // toothcnt overflowed to 0: like 57 - 0
	// TODO: review. With 15-1 and searchedtooth=27 (or 28), toothcnt=0 will fire FAKE !
	// This might not be a problem, since it already fired at 27 (or 28),
	// (at very worst we get a latefire anyway) and did step_ignumsc() step_ignumdc()
        if( allowedtooth >= (total_tooth - 3)){
                return 1;
        }
        return 0;
}
*/

// prefer access via macros for future flexibility
#define engine_tps255() engine.tps

// max value of engine.rpm_trigger_reenable

#define engine_rpm8() engine.rpm

#define engine_tps8() engine.tps


#define engine_rpm8() engine.rpm

#define engine_tps8() engine.tps

#define TRIGGER_REENABLE_MAX 5000   // 5000*4us = 20.0 ms

/***********************************************************************/

// define the adc sampling period [divisor from 1250Hz]
// THIS SHOULD MATCH H2KDIVIDE in eventqueue.h (maybe make it automatic?)
#define ADC_PERIOD  6

// TIMER3 TOP=1023. TIMER1 could use the same TOP value,
// but a little different value might be nice on the supply.
// The same frequency and artificial offset between TCNT3 and TCNT1 might be the best?
// PWM_FREQUENCY = 16MHz / 15.984kHz - 1
#define PWM_FREQUENCY   1000

// n-1 seconds after leaving crankmode we can't reenter (unless rpm drops to 0)
#define CRANK_TIMEOUT 11

// TPS >= 70%
#define FLOOD_CLEAR 179

// these make it harder for the user to make stupid bugs!
#define COLD_TEMP 60     // -40 F == -40 C
#define WARM_TEMP 174   // 165 F == 74 C (assume the engine is warm at this temperature)
// #define COLD_TEMP (config_ext.clt_range[0]) // first clt index.
// #define WARM_TEMP (config_ext.clt_range[9]) // last clt index (assume the engine is warm at this temperature. The constant 74C is a better idea, because one sometimes uses the last bin to add some fuel at very hot coolant)

// calculation from battcal calibrated 'real' vbat ( batADC = vbat / 30 * 255 )
#define BATT_6V 51
#define BATT_7V 60
#define BATT_14V 119
#define BATT_19V 162

// roadspeed interrupt divider definitions:
#define ROAD_SPEED1_DIVIDER ( config.roadspeed_divider & 0x0F )
#define ROAD_SPEED2_DIVIDER ( config.roadspeed_divider >> 4 )

//baro mode defines (uses boostcontrol too)
#define BARO_MODE_BIT 7
#define BARO_MODE() (config.baro_mode & _BV(BARO_MODE_BIT))
#define BARO_INPUT() ((config.baro_mode & 0x70) >> 4 )
#define config_dbaro() ( (config.baro_mode & 0x0f) << 2 )

/***********************************************************************/
/* Configuration variables */
/***********************************************************************/

// DO NOT PUT IFDEFS here unless you teach stru.pl to parse my_make
// Otherwise magicstr will be offsetted, and you will have a bad day. You have been warned.
struct config_t {
  // F1
  uint8_t primep;               // Priming pulse at first clt bin
  uint8_t primep_temp_scaling;  // Priming pulse scaling  at last clt bin (80=half, ff=same as cold)
  uint8_t awe_time_cold;	// AWE decay time, cold engine [4 cycles]
  uint8_t awe_time_warm;	// AWE decay time, warm engine [4 cycles]; Avoid 224 (896): can be special debug mode
  uint8_t cranking_thres; // Cranking threshold [100 rpm]

  uint8_t warmup_map_scale;		// 219 default
  uint8_t cranking_div;			// cranking divider
  uint8_t film_coeff;         // film coefficient

  // F2
  uint8_t shiftcut_potlow; // shiftcutinput active if less than this voltage (default 5F)
  uint8_t shiftcut_pothi;  // shiftcutinput active if greater than this voltage (normally FF)
  uint8_t shiftcut_retcut; // 7:4 retard 0..60deg absolute, 3:0 igncut 0..F: 6..100% default:0F
  uint8_t boost_matdec; // boost decrease -4kPa/C above this MAT [ C-100 ]
  uint8_t boost_egtdec; // boost decrease -1kPa/C above this EGT [8C]
  uint8_t iacref_matslope; //3:0 MAT (0..70C) raises iacduty (F: max 13%)
  uint8_t shiftcut_minrpm; // shiftcut min RPM [40 RPM]
  uint8_t egt_rich;			// egt rich gain, slope 255 is 38.9% per 100deg
  uint8_t knock_rich;			// knock rich gain, slope 255 is 24.6%per 1deg
  uint8_t knock_boost;		// boost reduce kpa per 1deg, not used yet

  uint8_t warmup_clt[10];        // Warmup enrichment bins(fn temp), biased at 100
  uint8_t warmup_rpm_scale;	// Warmup enrichment scaler at highest rpm bin: FF=rpm independent

  // F3
  uint8_t req_fuel;     // Fuel constant (req_fuel)
  uint8_t divider;      // IRQ divide factor for pulse
  uint8_t alternate;    // Bit2:0 primary injector count, bit4 1:batch fire when cranking
  uint8_t injopen;      // Injector open time at 13.2V [16 usec]
  uint8_t battfac;      // Battery gamma factor [16 usec based]
  uint8_t kpafac;	// MAP range [2kPa] 0: use A/B lookup table
  uint8_t kpaoffs;	// MAP signal offset -64..+31.75kPa, 128 is 0kPa [500/250 Pa slope]
  uint8_t injocfuel;    // injector opening effective rampup time [24 usec]
  uint8_t injrampup_battfac; // injector rampup slope battery dependence (FF)
  uint8_t injpwm;       // Injector PWM duty cycle (when current limiting is active)
  uint8_t injpwmt;      // Injector PWM time at which to activate pwm [100 usec]
  uint8_t als_launch_input;	// ALS and launch activation input channel

//  uint8_t rpmk[2];      // UNUSED. reserved for N2O enrichment % at 2000 and 6000 RPM
	uint8_t launch_maxkpa;	// start removing retard/enrichment above this [2kpa]
	uint8_t launch_kparange; //no retard / no enrichment at launch_maxkpa + launch_kparange [2kpa]

  // F4
  uint8_t tpsdot_kpadot_conf;  // 0: TPSDOT, 1: MAPDOT acceleration enrichment
  uint8_t tpsdotrate[4];  // TPS/MAPDOT acceleration bins, dv/dt
  uint8_t tpsaq[4];       // TPS/MAPDOT acceleration amount (fn TPSDOT) [100 usec]

  uint8_t als_cut_mintps;   // ALS cut mintps
  uint8_t tpsasync;     // 0,1:free. bit6-2: TPS acceleration halflife -0.1 ./ log(a/64)(sec) bit7: als_tks_powered
  uint8_t acmult;       // Acceleration cold multiplication factor, biased at 100
  uint8_t tpsacold;     // Cold acceleration added amount (at first clt bin) [100 usec]

  uint8_t launch_enrich;   // launch, fuel enrichment [0.4%]
  uint8_t launch_retard;  // launch, fixed ign. retard [1/4 crankdeg]

  uint8_t overrun_fuelcut;        // Overrun fuelcut, injectors disabled above this [100 rpm]
  uint8_t overrun_fuelresume;     // Overrun fuelresume, injectors reenabled below this [100 rpm]

  uint8_t rev_limit;      // Rev limit (hard) [100 rpm]

  // F5
  uint8_t airden_ignore;  // Fake airdensity low limit (low rpm only), 0==disable

  uint8_t config11;     // c11, bit7-4:ncyl-1,bit3:ALS over launch bit2:2ndWBO2 bit1: injector battery compensation strategy bit0: fix_retard_strategy
  uint8_t config12;     // c12, bit7:HALLdirac filter primtrig,bit6:HALL dirac filter sectrig,bit5:AIM Speedsensor div10,bit4,3:iac integral modificators(see below),bit2: alphan baro, bit1: PWM fuelpump, bit0: altRT
  uint8_t config13;     // c13, bit7, mat_type bit 6: clt_type, bit5:ign,lambda tpskpa bit4:MATretard disable bit3:baro bit2:fuel strategy no load multiplication bit1:WBO2 bit0:oddfire

  uint8_t batt_cal;	// VBatt calibration appr.BC=default
  uint8_t iac2load;     // alphaN only, iac contribution to load

  uint8_t als_max_cut;    // ALS max cut
  uint8_t baro_mode;          // max difference in barometric pressure

  uint8_t tps_low;        // tps, min adc-count
  uint8_t tps_high;       // tps, max adc-count

  uint8_t fan_temp;       // coolant fan, start temperature
  uint8_t fan_hyst;       // coolant fan, hysteresis
  uint8_t fan_channel;	  // coolant fan, output selection

  // IAC stuff F6
  uint8_t iac_step_seq;        // iac, stepping sequence (define iac wiring)
  uint8_t iac_conf;            // iac, bit7:softidle bit6:asym bit5:dualsol bit4=1:ignadv ON, bit3:preciseidle bit2:stepper(0=pwm) bit1:tps_follow bit0:power_iac_only_when_move
  uint8_t iac_max_steps;       // iac, max steps to extend
  uint8_t iac_tps_thres;       // iac, engine.tps value for idle threshold
  uint8_t iac_cold_idle_temp;  // iac, max temperature for cold idle speed
  uint8_t iac_warm_idle_temp;  // iac, min temperature for warm idle speed
  uint8_t iac_cold_rpm;        // iac, idle rpm (cold) [10 rpm]
  uint8_t iac_warm_rpm;        // iac, idle rpm (warm) [10 rpm]
  uint8_t iac_cold_start_pos;  // iac, crank position at first clt bin
  uint8_t iac_warm_start_pos;  // iac, crank position at last clt bin
  uint8_t iac_afterstart_rpm;  // iac, afterstart rpm [10 rpm], added to idle rpm
  uint8_t iac_afterstart_duration; // iac, duration of afterstart rpm [0.1 sec]
  uint8_t iac_speed; 		// iac, actuator 1/frequency [0.4 msec] (pwm/stepper:+4/+1)
  uint8_t iac_kp;              // iac, P-term
  uint8_t iac_ki;              // iac, I-term
  uint8_t iac_kd;              // iac, D-term
  uint8_t launch_rpm;  // Launch RPM [100 RPM]
  uint8_t iac_integral_limit_dec;  // iac, integral decrease limit
  uint8_t iac_integral_limit_inc;  // iac, integral increase limit
  uint8_t iac_integral_deadband; // iac, integral deadband [1 rpm] (being tested for usefulness)
//  uint8_t iac_deadband;        // UNUSED. was:iac, deadband [1 rpm]
	uint8_t launch_retard_margin; // full launch_retard above launch_rpm - launch_retard_rpm [4rpm]
  uint8_t als_cut_maxtps;        // iac, cut maxtps
  uint8_t roadspeed_divider;		// upper 4bit divide roadspeed2 input lower 4 bit divide roadspeed1 input
  uint8_t iac_ref_pos[10];     // iac, reference position (fn temp)
  uint8_t iac_sol_channel;	// iac output channel
  uint8_t iac_ign_slope; // iac, ignition advance [0.25deg/256rpm]
  uint8_t softrpm_range;  // Ignition SoftRPM range [4 RPM]
  uint8_t iac_ign_advance_limit;  // iac, ignition advance limit [0.25deg]
  uint8_t iac_ign_retard_limit;   // iac, ignition retard limit [0.25deg]
  uint8_t iac_ign_threshold;      // iac, threshold for activating air control [0.25deg]

  // ego controller F7
  uint8_t ego_conf;        // ego, bit6:LEAVE_ENRICH, bit5:NBO2_adc7 bit1:LSU4nowarmuponboot bit0:PID

  uint8_t ego_lag;         // ego, timeconstant (engine cycles) for changing correction factor
  uint8_t ego_coolant;     // ego, min coolant temperature for ego enabling[F]
  uint8_t ego_maxtps;      // ego, max tps position for ego correction
  uint8_t ego_maxmap;      // ego, max map for O2 correction[kPa]
  uint8_t ego_minrpm;      // ego, min rpm for ego enabling [100rpm]
  uint8_t ego_maxrpm;      // ego, max rpm for O2 correction [100rpm]
  uint8_t ego_warmup;      // ego, warmup time [sec]
  uint8_t ego_lean_limit;  // ego, max percent to lean afr[0.4%]
  uint8_t ego_rich_limit;  // ego, bit7:free, bit6:0 max percent to rich afr[0.4%]

	// note that NBO2 EGO PID reuses all the 5 values below
  uint8_t ego_pid_kp;      // ego, incredible speed limit (or pid kp)
  uint8_t als_enrich;	// ALS enrichment
  uint8_t ego_delta;       // ego, step size (percent) [0.4%]
  uint8_t ego_target;      // ego, target voltage (NBO2)
  uint8_t ego_pid_window;  // ego, minimum window for pid to react

  uint8_t ignretard_clamp;     // ignition retard claming.0=clamp at 63.75 (no clamping); 200=clamp at 64-(200/4) ATDC;
  // WBO2 F8
  // heater:
  uint8_t als_react_tps;   // ALS tps threshold
  uint8_t wbo2_abs_limit;	// wbo2, heater absolute limit [2a=12V ff=30V]
  uint8_t wbo2_limit_maxt;     // wbo2, max time spent at abs_limit [262msec]
  uint8_t wbo2_fallback;     // wbo2, fallback V after maxt [1c=10V ff=30V]
  uint8_t wbo2_retry_t;     // wbo2, retry control after [4sec]

  uint8_t als_retard_rpm1;	// ALS, retard RPM1 [100 rpm]
  uint8_t als_retard_rpm2;	// ALS, retard RPM2 [100 rpm]
  uint8_t wbo2_ri_target;	// wbo2, ri_target
  uint8_t wbo2_nernstdc_target; // wbo2, target nernstDC [5V/256]
  uint8_t wbo2_pump_pw_zero;    // wbo2, pump zero dutycycle (default: 66) [5V/256]
  uint8_t wbo2_calibration;     // wbo2, calibration constant of SENSOR
  uint8_t wbo2_heater_pid_kp;   // wbo2, heater pid kp
  uint8_t wbo2_heater_pid_ki;   // wbo2, heater pid ki
  uint8_t wbo2_heater_pid_kd;   // wbo2, heater pid kd
  uint8_t wbo2_heater_pid_ilimit;   // wbo2, heater pid integral limit

  uint8_t wbo2_pump_pid_kp;   // wbo2, pump pid kp
  uint8_t wbo2_pump_pid_ki;   // wbo2, pump pid ki
  uint8_t wbo2_pump_pid_kd;   // wbo2, pump pid kd (always 0)
  uint8_t wbo2_pump_pid_ilimit;   // wbo2, pump pid integral limit

  uint8_t inj_stage2_offset;	// injector staging offset 16 usec (-2048 .. +2048 usec)

  // Knock sampling F9
  uint8_t knock_chan;		// knock channel for each cyl; F0=disable
  uint8_t knock_sampling_window;   // knock, sampling window [4 degree] length:start(ATDC)

  uint8_t knock1_frequency;    // knock1, bandpass center frequency
  uint8_t knock1_gain;        // knock1, gain
  uint8_t knock1_integrator;  // knock1, integration time constant

  uint8_t knock2_frequency;    // knock2, bandpass center frequency
  uint8_t knock2_gain;        // knock2, gain
  uint8_t knock2_integrator;  // knock2, integration time constant

  // Knock action
  uint8_t knock_threshold;       // knock, threshold
  uint8_t knock_noise_scale;     // knock, noise scale factor (0x10= *1)
  uint8_t knock_max_retard;      // knock, max ignition retard [1/4 degree]
  uint8_t knock_default_retard;  // knock, default ignition retard [1/4 degree]
  uint8_t knock_retard_step;     // knock, ignition retard step [1/4 degree]
  uint8_t knock_retard_delay;    // knock, ignition retard delay [15ms]
  uint8_t knock_advance_step;    // knock, ignition advance step [1/4 degree]
  uint8_t knock_advance_delay;   // knock, ignition advance delay [15ms]
  uint8_t knock_minrpm;          // knock, min rpm [100rpm]
  uint8_t knock_maxrpm;          // knock, max rpm [100rpm]

  uint8_t ego_delay;			// EGO reactivation delay, cycles
  uint8_t ego_minmap;      // ego, min map for O2 correction[kPa]
  uint8_t cold_ignadjust;    // ign adjust at 0degC, for faster warmup[0.25deg]-128
  uint8_t crank_ve;    // VE value for cranking
  uint8_t film_enlean;    // Enlean downscaling
  uint8_t iac_close_delay; //delay before zero integral, and start closing solenoid-valve(if not tps_follow_mode configured)
  uint8_t iac_integral_lowmap; //lower kPa-limit, don't use integral control under this limit
  uint8_t barofac; //barometric sensor calibration factor
  uint8_t baro_offs; //barometric sensor calibration offset

  uint8_t lcd_c0;		 // LCD flags 7: leds 6: disabled 5: sw brightness 4: pwm brightness 3:reinit 2:aim 1:lcd_easygoto 0:busypolled
  uint8_t lcd_delay;		 // LCD-cablelength delays Higherhalf:beforeENable LowerHalf:ENable
  uint8_t lcd_backlight;	 // LCD backlight: FF is fully on

  uint8_t kpa_conf;		 // bit7-4: kpafac_unit, bit3-0: kpa_config_resolution
  uint8_t als_igncut_rpm2;	 // ALS igncut RPM2 [100 rpm]
  uint8_t als_retard_mintps;	// ALS retard mintps[0.4%]
  uint8_t als_retard_maxtps;	// ALS retard maxtps[0.4%]

  uint8_t lcd_default_view;      // LCD, default runtime view

  // F11
  uint8_t primary_trigger;      // trigger1, bit0 0:falling/1:rising, bit1 0:missingtooth/1:coil, bit2=1:filtering, bit3:advanced multitooth, bit4:missing N-1, bit5:FiatStilo, bit6:AudiTrigger135+ bit7:audionly 0=divideby3, 1=divideby2
  uint8_t secondary_trigger;    // trigger2, bit0 0:falling/1:rising, bit1 0:enable/1:disable, bit2 1:filter, bit3 0:toothwheel/1:coil, bit4 0:alien advance/1:cam sync, bit5 1:both edge at cranking bit6:fiatstilo

  uint8_t tooth_wheel;          // tooth wheel, number of teeth on the wheel
  uint8_t trigger_tooth;        // tooth wheel, active trigger tooth (00:missing tooth)
  uint8_t another_trigger_tooth; // tooth wheel, define another trigger tooth
  uint8_t lcd_analin;		// lcd stepping button input

  uint8_t tooth_wheel_twidth[2];         // toothwidth [0.25 crankdeg]

  uint8_t cam_sync_r_edge_phase;      // not used! cam sync, rising edge, engine phase at next primary trigger
  uint8_t cam_sync_f_edge_phase;      // not used! cam sync, falling edge, engine phase at next primary trigger

  uint8_t reset_engphase_after; // not used!

  uint8_t ign_tdcdelay;		// TDC after the trigger(0.5crankdeg)
  uint8_t ign_dwell14;		// base dwell above 14V (64usec)
  uint8_t ign_dwellmap;		// added MAPdwell slope (00: dwell14 used, VBATT not applied either)
  uint8_t ign_crank_advance;    // ignition advance at cranking [0.25crankdeg]
  // developer bit-definition of ign_out:
  // 7:disable, 6:EDIS/dummy, 5:after/before, 4:ref_to_trigger/ref_to_spark 3:multispark 1: ign_dual_out
  // EDIS bits will be	0000...?
  // dummy-ign bits are:0111...?
  // disable ign:		1111....
  uint8_t ign_out;	// EDIS:0x0? dummy: 0x7? disable:0xff bit1: ign_dual_out bit0: invertout
  uint8_t ignchmax;  // h[2] maxindex for ignition channel lookup

  // Fuelpump section:
  uint8_t engine_off_delay;		// turn off fuelpump after inactivity time [262msec]
  uint8_t pump_on_mintime;		// minimum time to leave fuelpump on at startup [262msec]
  uint8_t fuelpump_channel;		// fuelpump output channel

  // Injector staging section F12
  uint8_t inj_stage2_rate;		// secondary injectors rate to calculated pw [0.4%]
  uint8_t inj_stage2_start_tps;		// secondary injectors open above given throttle
  uint8_t inj_stage2_start_map;		// secondary injectors open above given manifold pressure

  // AntiLagSystem section
  uint8_t als_igncut_rpm1;		// als igncut RPM1 [100 rpm]
  uint8_t als_deact_time;		// als,switchoff time [sec]
  uint8_t als_deact_egt;		// als,switchoff egt [8 deg celsius]
  uint8_t als_retard;			// als,max ign. retard [1/2 crankdeg]
  uint8_t als_rev_limit;        // Rev limit when ALS active [100 rpm]

  // Misc power outputs
  uint8_t misc1out_minrpm;		// misc1out,minimum rpm [100 rpm]
  uint8_t misc1out_maxrpm;		// misc1out,maximum rpm [100 rpm]
  uint8_t misc1out_mintps;		// misc1out,minimum tps position
  uint8_t misc1out_maxtps;		// misc1out,maximum tps position
  uint8_t misc1out_minmap;		// misc1out,minimum manifold pressure
  uint8_t misc1out_maxmap;		// misc1out,maximum manifold pressure
  uint8_t misc1out_channel;		// misc1out,output channel selector

	uint8_t misc1out_conditions;// bits 0-3: condition one, 4-7: condition two
	uint8_t misc1out_condition1;// misc1out condition1 value
	uint8_t misc1out_condition2;// misc1out condition2 value
    uint8_t secignore;//ignore secondary trigger pulses which tooth_cnt bigger then secignore, default 0xff
    uint8_t gps_log;// GPS logging to SD card
  // Misc power outputs
    uint8_t iac_close_minmap; // minimum map value to force iac to iac_close_pos
    uint8_t iac_close_pos;
  uint8_t act_wot_rpm;		// WOT switch RPM threshold (FF=disable)
  uint8_t act_wot_channel;	// WOT switch output channel
  uint8_t act_rpm_rpm;		// RPM switch RPM threshold (FF=disable)
  uint8_t act_rpm_channel;	// RPM switch output channel
  uint8_t egt1_cal;			// EGT calibration multiplier
  uint8_t als_tks_channel;		// ALS TKS output channel
  uint8_t boost_conf;	// boostcontrol bit7:4 PWM period [8msec]; bit3: MAT/EGT reduces refpwm bit2:gear bit1:wheelspeed bit0 mode 0 relative, 1 absolute
  uint8_t boost_more; // bit7:targetoffs 0=0,1=100kPa; bit6..5 00=MAPtarget, 01=RPMtarget,10=pgastarget,11=boosttarget;bit3:0 = boost alternate PWM period
  uint8_t boost_minpressure;	// actuator OFF below this pressure
  uint8_t boost_pid_kp;   // boostcontrol pid kp
  uint8_t boost_pid_ki;   // boostcontrol pid ki; reused by als; see fuelcalc.h
  uint8_t boost_pid_kd;   // boostcontrol pid kd;
  uint8_t boost_pid_ilimit;   // boostcontrol pid integral limit

  uint8_t boost_channel;	// boostcontrol solenoid output channel

  // water pump control
  uint8_t water_pump_temp;	// water pump, start temperature

  uint8_t camshaft_ex_channel;	// camshaft angle control output, exhaust side
  uint8_t nonve_kpalookup_minkpa;	// for ignadv,lambdatarget etc... TPS lookup used below this [2kPa
  uint8_t water_pump_hyst;	// water pump, hysteresis
  uint8_t water_pump_channel;	// water pump, output channel

  uint8_t toothrel_normal;	// Nn: tg_prev * n/16 <= tg <= tg_prev * (1+N/16)
  uint8_t toothrel_missing;	// Mm: tg_prev * (1.25 + m/8) <= tg <= tg_prev * (2.5 + M/8)

  // fuelcuts
  uint8_t fuelcut_min_kpa;	// fuelcut below this MAP (00:disable; 10 recommended)
  uint8_t fuelcut_max_kpa;	// cut the fuel if the engine kpa exceeds this value 0xFF=disable (even if selecting alpha-n)

  // tach output control
  uint8_t tach_channel;         // tach signal output channel FF=disabled
  uint8_t tach_div;      	// divide tach freq by X/16; default 16. 1-5 changes duty instead of freq

  // full power shift cut variables
  uint8_t shiftcut_conf_input;		// bit6:revcut spark,bit5:shiftcut spark,bit4:shiftcut fuel,3:0 mcp3208ch
  uint8_t shiftcut_tpsthresh;	// min TPS to allow shiftcut (default: 0)
  uint8_t shiftcut_time; 		// igncut time after inputpulse [4 msec]
  uint8_t softrpm_rpm;		// Ign Softrpm RPM [100 RPM]
};


/** config_ext variables
* Max size should be no more then 248 byte, please!
See sipr_protocol.c!
*/
struct config_ext_t {
    // n2o activation
  uint8_t n2o_mintps;           // minimum TPS for activation (full throttle switch)
  uint8_t n2o_minrpm;           // minimum RPM for N2O activation
  uint8_t n2o_maxrpm;           // maximum rpm, switch off n2o above this rpm (recommended to hit this before revlimiter)
  uint8_t n2o_retard;           // amount of timing retard when function activated
  uint8_t n2o_minkpa;           // minimum MAP for activation (backup for TPS signal, we dont want to activate on too low load, will cause explosion)
  uint8_t n2o_maxkpa;           // maximum MAP for activation (some people use N2O only to help turbos spool quickly)
  uint8_t n2o_input;            // input channel for activation
  uint8_t n2o_output_channel;   // output channel to activate solenoid
  uint8_t n2o_settings;         // bit0:0=disabled,1=enabled, bit1=1:launch control switch deactivates n2o 1=on(turbo guys might want N2O to spool turbo's before launch, N/A guys wants to activate nitrous when leaving starting line), bit2:shiftcut deactivates n2o
	// Andrey's suggestion (not implemented)
	// bitN: switch tables when active
	// bitN: share max egt with warning config
	// bitN: share max lean with warning config
	// bitN: share max clt with warning config
	// bitN: knock disables n2o
    // launch stuff
    uint8_t launch_out_channel;   // outputchannel active when launch (button?) active
    uint8_t softcut_rpm1;       //rpm limit[64RPM]
    uint8_t softcut_rpm2;
    uint8_t softcut_time1;      //time limit[8ms]
    uint8_t softcut_time2;
    uint8_t softcut_cut;        //lower 4 bit: cut1: 6.25..100% higher 4 bit cut2:6.25..100%

    uint8_t launch_out_conf;      // bit0: launch_out_conf_enable
                                  // bit1: launch_out_conf_pwmramp, disable/enable pwm ramping function Can be replaced by possibility to configure ramp from 0-100% in 0 mS, if 0% is fully off for sure, and 100% is fully on for sure
				  // bit3: 0=flameoff(not removing fuel when launch MAP too high) 1=flameon (removing launch extra fuel bangs in exhaust)
                                  // bits 4:7 launch_out_conf_pwmfreq, same as boost_conf_pwm
	uint8_t iac_minpwm;   // iac minimum PWM (audi/vw limphome iac ~24%, opel/gm ~32%). Also applied for stepper. Default:0
	uint8_t boostgauge_channel; 	// Channel for boostgauge, 0x16 on audi
	uint8_t boostgauge_offset;	// Boostgauge offset, 43 on audi
	uint8_t boostgauge_mul;		// Boostgauge multiplier, 71 on audi

	uint8_t map_filter;         // map filter settings
						//[7..4] - number of triggers, [3..0] - minimum samples num for average
	uint8_t speed_sensor;		// speed sensor calibration, freq on 100kph
	uint8_t gears[6];			// speed for 5000rpm
	uint8_t pwcommlock0;		// 0xde (222) applies communication lock (password needed to use comm)
	uint8_t als_iacpos;		// iac position when ALS active (0 disables)
	uint8_t anytrim2_in; //  bit3:use MAP instead of mcp3208, 2:0 mcp3208 channel
    uint8_t anytrim3_in; //  bit3:use MAP instead of mcp3208, 2:0 mcp3208 channel
	uint8_t strig_validation;	// secondary trigger validation bit 7: filter 6:shutdown 5:ls_1_forgiving 1..0: camshaft measure count for Subaru and Suzuki VVTI 
    uint8_t anytrim3_type;  // anytrim 3 mode:bit7:4 (0=fueltrim), 1=lambdatrim, 2=spark adjust, (4=boost trim mult), (5=boost trim add) F=disable;
	uint8_t speed_mintime; //mintime for speedsensor period [64us]
    uint8_t softcut_enrich; // at soft fuel cut we can enrich fuel too for partial fuel cut from 100% ( 0 ) to 147%(15),lower 4 bit is for low RPM cut, higher 4 bits for high RPM cut
    uint8_t vvti_measure_tooth; //normal ( missing tooth + vanos ) cam angle measure tooth_cnt on 4 bits, upper third trig; lower sectrig
	uint8_t max_prim_per_sec_count; //maximum primary pulse per sectrig pulse for sectrig filtering
	uint8_t sd_logging_mode;	//  bit 7..4 log time after condition disables; bit 3..0 logging start condition  (0: always 1: misc1 2: misc2 3..15 unused)
	uint8_t boost_pwm_max;	// maximum value for boost pwm
	uint8_t boost_pwm_min;	// minimum value for boost pwm
	uint8_t iac_conf2;	// iac, bit7: warning as CE, bit6: speedsensor both edges, bit5: 2nd speedsensor, bit4:wheelerr warning, bit2: float_stepper, bit1:toyota stepper at p259_4-7, bit0:stepper polarity

	uint8_t warning_rpm;		// active above rpm [100rpm]
	uint8_t warning_kpa;		// active above kpa [2kpa]
	uint8_t warning_conf;		// 7-6: skip time after fuelresume[0.2s], 5:knock,4:adc dir, 3-0:adc channel
	uint8_t warning_adc_value;	// adc threshold value, fuel pressure etc
	uint8_t warning_egt;		// max egt [8deg]
	uint8_t warning_afr;		// max lean difference with afr table [0.1afr]
	uint8_t warning_clt;		// max clt [-100C]
	uint8_t warning_channel;	// output channel

	uint8_t anytrim2_type;	// anytrim 2 mode:bit7:4 0=fueltrim, 1=lambdatrim, (2=spark adjust), 4=boost trim mult, 5=boost trim add F=disable;
	uint8_t boost_tps_dc_scale;	// scale for dc from target(tps) table, 64=100%
	uint8_t n2o_mingear;        // minimum gear for enable

    uint8_t cylinder_order[3];//used just for VT, need to move to eeprom only

	uint8_t als_tps_thres;	// TPS threshold when ALS switch active
    uint8_t strig_min_angle;	// secondary trigger validation min angle [4deg]
	uint8_t camshaft_ex_refpos;	// camshaft exhaust reference position

	uint8_t knock_gain[8];		// gain for each cyl
	uint8_t knock_reference[8];	// noise reference value for each cyl
	uint8_t knock_calibration_min;	// minimum noise for selfcalibration interval
											// 7bit: selfcalibration enabled
	uint8_t knock_calibration_max;	// max noise for selfcalibration interval

    uint8_t inj_stage2_start_rpm; // inj staging disabled below this RPM
    uint8_t transb_config;  // bit 7..4: transb_frequency: 1-16Hz bit 3..0: transb_pulses: Number of pulses:0-15 (0 disables)
    uint8_t transb_input;   // bit 3..0: transb button analog input  bit 7..4:unused
    uint8_t transb_rpm1;    //transb rpm range1
    uint8_t transb_rpm2;    //transb rpm range2
    uint8_t transb_dwell1;  //transb dwell for rpm1
    uint8_t transb_dwell2;  //transb dwell for rpm2
    uint8_t transb_channel; //transb output channel

	uint8_t inj_deadtime[5];	// inj deadtime table 8,10,12,14,16
	uint8_t speed2_sensor;// second speedsensor calibration
	uint8_t fp_divider;// FP PWM rate
	uint8_t gears_delay;// experimental for launch, excluded from release
	uint8_t anytrim1_in; // bit3:use MAP instead of mcp3208, 2:0 mcp3208 channel
	uint8_t launch_speed1;// experimental, excluded from release

	uint8_t n2o_rich; // fuelcorr for n2o, 256:+50% 128:0%, 0:-50%
	uint8_t anytrim1_type; //bit 7:4 = anytrim 1 function, F for disable , bit 3:0 unused
	uint8_t accell_retard; //retard for accell=100%*reqfuel
	uint8_t cranking_clt_range[4];	// cranking and afterstart coolant bins
	uint8_t crank_enrich[4];	// Cranking enrichment
	uint8_t afterstart_enrich[4];//afterstart enrichment

///the following 7 bytes also used for camshaft angle control
	uint8_t vvti_cold;			// force campos for cold engine or pwm period for camshaft angle(bit4..7), third interrupt enable (bit3), dual output control (bit2)
	uint8_t vvti_idle;			// force campos for idle or reference position to camshaft angle
	uint8_t vvti_pid_kp;		// P term
	uint8_t vvti_pid_ki;		// I term
	uint8_t vvti_pid_kd;		// D term
	uint8_t vvti_pid_ilimit;	// I limit
	uint8_t vvti_channel;		// out channel, vvti actuator

    uint8_t accel_rpm[4];		// acceleration enrichment rpm switchpoints
    uint8_t accel_scale[4];		// acceleration enrichment scale

    uint8_t strig_max_angle;	// secondary trigger validation max angle [4deg]

    uint8_t misc2out_minrpm;		// misc2out,minimum rpm [100 rpm]
    uint8_t misc2out_maxrpm;		// misc2out,maximum rpm [100 rpm]
    uint8_t misc2out_mintps;		// misc2out,minimum tps position
    uint8_t misc2out_maxtps;		// misc2out,maximum tps position
    uint8_t misc2out_minmap;		// misc2out,minimum manifold pressure
    uint8_t misc2out_maxmap;		// misc2out,maximum manifold pressure
    uint8_t misc2out_channel;		// misc2out,output channel selector
	uint8_t misc2out_conditions;// bits 0-3: condition one, 4-7: condition two
	uint8_t misc2out_condition1;// misc2out condition1 value
	uint8_t misc2out_condition2;// misc2out condition2 value

    uint8_t inj_size[2];				// injector size cc per min
	uint8_t clt_curve[17];		// clt curve
	uint8_t mat_curve[17];		// mat curve
	uint8_t clt_range[10];		// clt bins
	uint8_t boostalt_channel;	// custom PWM output/ boost alternate output channel
	uint8_t sectrig_maxrpm;		// sectrig is disabled above this [100RPM]. 1F (3100 RPM) for Nissan Default: FF
/** config_ext variables
* Max size should be no more then 248 byte, please!
See sipr_protocol.c!
*/
};

struct at_config_t
{
	uint8_t shift_up[3][8];//shift up speed for each gear, [0] - 1 to 2 gear
	uint8_t shift_dn[3][8];//shift down speed for each gear, [0] - 2 to 1 gear
	uint8_t lockup_lock[8];//lock speed
	uint8_t lockup_unlock[8];//unlock speed
	uint8_t tb_table[8];//tps bins
	uint8_t sol1_table[4];//sol1 position for each gear see true table for gears
	uint8_t sol2_table[4];//sol2 position for each gear see true table for gears
	uint8_t wot_shift[3];//WOT-mode rpm shift points table for max performance
	uint8_t wot_speed[3];//WOT-mode downshift speed
	uint8_t wot_threshold;// 90% default, WOT-mode activation
	uint8_t wot_hyst;// 1% default
	uint8_t options;//see notes below
	uint8_t lock_gear;// lowest gear to allow lock
	uint8_t lock_delay;//shifting time
	uint8_t retard;//retard angle, 0 - no retard, not used in WOT mode if shift-cut enabled
	uint8_t retard_time;//time for retard, not used in WOT mode if shift-cut enabled
	uint8_t tc_kpa_unlock;//force unlock TC above pressure to prevent lock-up damage under hi-load condition

	uint8_t sol1_channel;//shift solenoid 1
	uint8_t sol2_channel;//shift solenoid 2
	uint8_t tcsol_channel;//lock-up solenoid

	uint8_t drive_channel;//resistor divider circuit from selector
	uint8_t dyno_gear;//0-dyno mode disabled, !=0 force select gear for dyno
	uint8_t brake_channel;//brake switch
	uint8_t pressure_pwm[8];//tps2pwm transfer table
	uint8_t pwm_channel;//tps2pwm output
	uint8_t tc_kpa_lock;//lock TC only below pressure to smooth lock-up
	uint8_t idle_rpm_up;//idle up rpm for drive mode
	uint8_t idle_ref_up;//idle up steps for drive mode
	uint8_t neutral_channel;// 0V - N,P, 5V - other
	uint8_t at_temp_curve[17]; // oil temp sensor curve
	uint8_t at_temp_channel; // ADC channel, oil temp
};

//prohibit_config:
#define PROHIBIT_CONF_IBUTTON 7 // bit7 ibutton or version check configured
#define PROHIBIT_CONF_IBUTTON_KEEPALIVE 6 // use adc to keep alive
#define DISABLE_PROHIBIT_H  0xBE    //equal to ver. 7.5.128
#define DISABLE_PROHIBIT_L  0xEF
struct config_shared_t {
    uint8_t prohibit_config;//injector and ignition prohibiting and ibutton configuration bits see above
    uint8_t prohibit_h;     //prohibiting(version) or ibutton value high byte
    uint8_t prohibit_l;     //prohibiting(version) or ibutton value low byte
    uint8_t config_switch;  //table switch
	uint8_t vtstorage[7];	// storage for VT 
	uint8_t blab[8]; // storage for VT, the name is totaly bad cant change
	uint8_t fix_retard[5];//gear / speed dependent fix spark retard
	uint8_t wbo2_pump_pw_zero2; // wbo2 2nd channel
	uint8_t wbo2_nernstdc_target2; // wbo2 2nd channel
	uint8_t wbo2_calibration2; // wbo2 2nd channel moved from config !

/** config_shared variables
* Max size should be no more then 248 byte, please!
* EEPROM storege adresses are hardcoded to this limit too.
See sipr_protocol.c!
*/
};
extern struct config_t config;
extern struct config_ext_t config_ext;
extern struct config_shared_t config_s;


#define KNOCK_SIZE_RPM 8
#define KNOCK_SIZE_KPA 8
#define KNOCK_SIZE (KNOCK_SIZE_RPM*KNOCK_SIZE_KPA)

#ifdef KNOCK_ALTERNATIVE
struct knock_noise_t {
	uint8_t knock_noise[KNOCK_SIZE];  // knock noise
	uint8_t rpm_range[KNOCK_SIZE_RPM];  // VE table RPM bins for 2-D interpolation
	uint8_t kpa_range[KNOCK_SIZE_KPA];  // VE table MAP bins for 2_D interpolation
};
extern struct knock_noise_t knock_table;
#endif

#ifdef TOYOTA_1JZ_VVTI
struct vvti_map_t {
	uint8_t vvti[IGNADV_SIZE];  // camshaft angle
};
extern struct vvti_map_t vvti_map;
#endif

#ifdef CHARGE_TEMP
struct f_charge_t {
	uint8_t f_charge[VE_SIZE_RPM*VE_SIZE_KPA];
};
extern struct f_charge_t f_charge;
#endif



struct table_shared_t {
    struct config_shared_t config_s;
#ifdef KNOCK_ALTERNATIVE
    struct knock_noise_t knock_table;
#endif
#ifdef CHARGE_TEMP
    struct f_charge_t f_charge;
#endif
#ifdef TOYOTA_1JZ_VVTI
    struct vvti_map_t vvti_map;
#endif

};

// how many entries in the VE-table
#ifndef VE_SIZE_RPM
#define VE_SIZE_RPM 8
#endif

#ifndef VE_SIZE_KPA
#define VE_SIZE_KPA 8
#endif

// check code when changing (better: do not change)
#define LAMBDA_SIZE_RPM 8
#define LAMBDA_SIZE_KPA 8
// size withOUT rpm,kpa ranges:
#define LAMBDA_SIZE (LAMBDA_SIZE_RPM*LAMBDA_SIZE_KPA)

// check code when changing (better: do not change)
#define MAP_CORR_SIZE_RPM 8
#define MAP_CORR_SIZE_KPA 8

#ifdef CAMSHAFT_ANGLE
#define CAMSHAFT_SIZE_RPM   8
#define CAMSHAFT_SIZE_KPA   8
#define CAMSHAFT_ANGLE_SIZE  ( CAMSHAFT_SIZE_RPM * CAMSHAFT_SIZE_KPA )
#endif

/**
 * We could have all lookup tables in one struct ?
 * It would make storage simpler and shorter, with no disadvantages.
 */
struct ve_t { // keep the order table, rpm_range, kpa_range !!!
  uint8_t ve8_table[VE_SIZE_RPM*VE_SIZE_KPA]; // main VE table
  uint8_t rpm_range[VE_SIZE_RPM];  // VE table RPM bins for 2-D interpolation
  uint8_t kpa_range[VE_SIZE_KPA];  // VE table MAP bins for 2_D interpolation

	uint8_t lambda[LAMBDA_SIZE];  // lambdacorr table
	uint8_t lambda_rpm_range[LAMBDA_SIZE_RPM];
	uint8_t lambda_kpa_range[LAMBDA_SIZE_KPA];

	uint8_t mat_retard[8*8]; // MAT/TPS retard table (0.25 crankdegree)
	uint8_t mat_range[8];	// MAT bins
	uint8_t tps_range[8];	// RPM bins

#ifdef MAP_CORR
	uint8_t map_corr[MAP_CORR_SIZE_RPM*MAP_CORR_SIZE_KPA];
	uint8_t map_corr_rpm_range[MAP_CORR_SIZE_RPM];  // VE table RPM bins for 2-D interpolation
	uint8_t map_corr_kpa_range[MAP_CORR_SIZE_KPA];  // VE table MAP bins for 2_D interpolation
#endif

#ifdef MAT_ENRICH
	uint8_t mat_enrich[8*8]; // main MAT/TPS enrichment table
	uint8_t emat_range[8];  // MAT table MAT bins for 2_D interpolation
	uint8_t etps_range[8];  // MAT table RPM bins for 2-D interpolation
#endif
#ifdef CAMSHAFT_ANGLE
    uint8_t camshaft_angleI[CAMSHAFT_ANGLE_SIZE];
//    uint8_t camshaft_rpm_rangeI[CAMSHAFT_SIZE_RPM];
//    uint8_t camshaft_kpa_rangeI[CAMSHAFT_SIZE_KPA];
    uint8_t camshaft_angleE[CAMSHAFT_ANGLE_SIZE];
    uint8_t camshaft_rpm_range[CAMSHAFT_SIZE_RPM];
    uint8_t camshaft_kpa_range[CAMSHAFT_SIZE_KPA];
#endif
	uint8_t launch_rpmtable[8*2];	// launch RPM in function of wheelspeed

};
struct ve_t ve;
#ifdef KNOCK_ALTERNATIVE
struct knock_noise_t knock_table;
#endif

#ifdef CHARGE_TEMP
struct f_charge_t f_charge;
#endif

#ifndef NOIGN
// check code when changing (better: do not change)
#define IGNADV_SIZE_RPM 12
#define IGNADV_SIZE_KPA 12
#define IGNADV_SIZE (IGNADV_SIZE_RPM*IGNADV_SIZE_KPA)

// currently ignition shares the same bins as ve table (easier on the tuner person)
struct ign_t {
	uint8_t adv_table[IGNADV_SIZE];  // ignition advance table [0.25 deg]
	uint8_t adv_table_rpm_range[IGNADV_SIZE_RPM];
	uint8_t adv_table_kpa_range[IGNADV_SIZE_KPA];
};
struct ign_t ign;
#endif

#ifdef BOOSTCONTROL
// take care if changing this
#define TPSTRANSFERTABLE_SIZE (8)

#define BOOST_DC_SIZE IGNADV_SIZE
#define BOOST_SIZE_KPA IGNADV_SIZE_KPA
#define BOOST_SIZE_RPM IGNADV_SIZE_RPM
#define BOOST_TPS_SIPRSIZE (TPSTRANSFERTABLE_SIZE + (BOOSTCONTROL_GEARS + 1) * BOOST_SIZE_RPM)
#define BOOST_REF_FULLTABLE (BOOST_SIZE_RPM*BOOST_SIZE_KPA)
// 2*6 lines for boost +1 line for rotary + inj_angle
#define BOOST_BTABLE_GEARLINES (BOOSTCONTROL_GEARS + BOOSTCONTROL_GEARS+2)

#define ANYTRIM_ARRAY_1 boost.anytrim1_array
#define ANYTRIM_ARRAY_2 (boost.anytrim1_array + 17)
#define ANYTRIM_ARRAY_3 (boost.anytrim1_array + 34)

#define ANYTRIM_PW      0
#define ANYTRIM_LAMBDA  16
#define ANYTRIM_SPARK   32
#define ANYTRIM_CRANK   48
#define ANYTRIM_BOOST_MUL 64
#define ANYTRIM_BOOST_ADD 80
#define ANYTRIM_BOOST_REF_ADD 96
#define ANYTRIM_LAUNCH 128
#define ANYTRIM_SOFTCUT 144

#ifdef INJ_ANGLE
#define INJ_ANGLE_CURVE         boost.inj_angle
#define INJ_ANGLE_RPM_RANGE     boost.boostgear_rpmrange
#define INJ_SIZE_RPM            BOOST_SIZE_RPM
#endif

#define AWE_TIME_WARM_DEBUG (config.awe_time_warm == 224)

struct boost_t {
// KEEP THE NEXT 6 lines together, because comm.c and menu.c TableDump depends on it !
// We could move these inside ve_t (or storage section) to save storage related flashspace
  uint8_t tpstransfer_table[TPSTRANSFERTABLE_SIZE]; //t[0] stored on same page as next, but dumped separately
  uint8_t boostmaxtarget_table[BOOSTCONTROL_GEARS * BOOST_SIZE_RPM]; // b[0..5] bottom 6 rows: RPM,gear1..6 => boostmaxtarget
  uint8_t boostpidrefpwm_table[BOOSTCONTROL_GEARS * BOOST_SIZE_RPM]; // b[6..11] RPM,gear1..6 => boostpid gear referencePWM or 2 lines for speed => refpwm
  uint8_t rotary_trailing_spark[BOOST_SIZE_RPM]; // b[c] rotary
  uint8_t inj_angle[BOOST_SIZE_RPM];  // b[d] injection angle table
  uint8_t boostgear_rpmrange[BOOST_SIZE_RPM]; // b[r] RPM bins for above tables
  uint8_t anytrim1_array[3*17]; // anytrim values, see anytrim function

  // uint8_t dummy_5_byte_table[17-BOOST_SIZE_RPM];//p table has 17 data per line we must allocate 5 more dummy bytes
// boostalt reference 'y' table:
union{
  uint8_t boost_ref_table[BOOST_REF_FULLTABLE];
#ifdef A340E
    struct at_config_t at_config_at;
#endif
};
	uint8_t boost_ref_rpm_range[BOOST_SIZE_RPM];
	uint8_t boost_ref_kpa_range[BOOST_SIZE_KPA]; // no +6 here !
};
struct boost_t boost;
#endif

#ifdef HWMAP
#define HWMAP_SIZE 8
#define HWMAP_LINES (6)
#define HWMAP_BYTES (HWMAP_SIZE * HWMAP_LINES)
// hardware mappings
struct hwmap_t {
	uint8_t table[HWMAP_BYTES];
};
struct hwmap_t hwmap;
#endif

#ifdef A340E
#define at_config boost.at_config_at
#endif

#ifdef TOYOTA_1JZ_VVTI
struct vvti_map_t vvti_map;
#endif

struct table_config_t {
    struct config_t config;
	struct config_ext_t config_ext;
    struct ve_t ve;
#ifndef NOIGN
    struct ign_t ign;
#endif
#ifdef BOOSTCONTROL
    struct boost_t boost;
#endif
#ifdef HWMAP
    struct hwmap_t hwmap;
#endif

};

#define shared_complement_size  (0xC0 - sizeof( struct table_shared_t))
#define table_complement_size  (0x780 - sizeof( struct table_config_t))
#define STOREDMAPCOUNT 1
#define SEEPROM __attribute__((section(".eeprom")))
extern struct table_shared_t table_shared SEEPROM;
extern uint8_t table_shared_complement_dummy[shared_complement_size] SEEPROM;
extern uint8_t Vemstune_message[64] SEEPROM;
extern struct table_config_t table_config_A SEEPROM;
extern uint8_t table_config_A_complement_dummy[table_complement_size] SEEPROM;
extern struct table_config_t table_config_multi[STOREDMAPCOUNT] SEEPROM;
extern uint8_t table_config_B_complement_dummy[table_complement_size] SEEPROM;

#define flamebit (config_ext.launch_out_conf & _BV(3))

#define actual_tps_thres (engine.status_ext & _BV( als_switch_activated ) ? config_ext.als_tps_thres : config.iac_tps_thres)

// should be in digitalout.h ?
#define output_disabled(ch) ((ch & 0xF) == 0xF)
#define output_enabled(ch) ((ch & 0xF) != 0xF)

#define CRANKING_CLT_RANGE_LEN 4
#define CLTRANGE_LEN 10
//extern uint8_t PROGMEM CLTRANGE[];
#define warmup_clt_range(X) config_ext.clt_range[X]
//#define warmup_clt_range(IDX) pgm_read_byte_near( CLTRANGE + (uint16_t) IDX)

#ifndef KPA_CONFIG_RESOLUTION
#define KPA_CONFIG_RESOLUTION (config.kpa_conf & 0x0F)
#endif

#ifndef KPAFAC_UNIT
#define KPAFAC_UNIT ((config.kpa_conf & 0xF0) >> 4)
#endif

#define FIATSTILO_SUBARU_MASK (_BV(ptrig_coil) | _BV(trigger_fiatstilo) | _BV(trigger_diveby2))
#define FIATSTILO_SUBARU ((config.primary_trigger & FIATSTILO_SUBARU_MASK ) == FIATSTILO_SUBARU_MASK)
// on Subaru boxer sectrig does not move when cam actuated, so measure tooth measures wheelspeed2 pulse instead of sectrig if:
#define SUBARU_TRIGGER4_ANGLE (config_ext.speed2_sensor == 1)

#define LANCIA_COSWORTH_MASK (_BV(ptrig_coil) | _BV(trigger_fiatstilo))
// the wider FIAT... mask is intentional (checking 3 bits):
#define LANCIA_COSWORTH ((config.primary_trigger & FIATSTILO_SUBARU_MASK ) == LANCIA_COSWORTH_MASK )
// under subaru6+7 (shortgaptrigger), some special cases, mostly for initial sync:
// coiltype && FIATSTILO_SUBARU && MASERATI  OLD was: (config.tooth_wheel == 6). Note (and strig_validation==00 ) => MASERATI
#define MASERATI (config.config13 & _BV(ODDFIRE))

// FIATSTILO_SUBARU && VIPER
#define VIPER (config.tooth_wheel == 10)

#define dirac_elimination_prim() (config.config12 & _BV(dirac_elimination_prim_bit))
#define dirac_elimination_sec() (config.config12 & _BV(dirac_elimination_sec_bit))
#define aim_speedsensor_mult10() (config.config12 & _BV(aim_speedsensor_divider_bit))
#define iac_integral_frozen()   (config.config12 & _BV(iac_integral_frozen_bit))
#define iac_integral_moving()   (config.config12 & _BV(iac_integral_moving_bit))
#define sd_bytestuffed()   (config.config12 & _BV(sd_bytestuffed_bit))
// k: pressure [0.25 kPa] (400 at 100kPa)
static inline uint8_t engine_kpa8_of(uint16_t k)
{
	uint8_t ret;
	if(KPA_CONFIG_RESOLUTION==2){
		// high-boost (50 at 100kPa)
		ret = ( k >= 255*8 ) ? 255 : k>>3;
	} else { // standard, low-boost: KPA_CONFIG_RESOLUTION==1
		// (100 at 100kPa)
		ret = ( k >= 255*4 ) ? 255 : k>>2;
	}
	return ret;
}

/* returns MAP in [1 or 2 kPa] unit, railed to max 255
* Variables compared against this value:
* config.inj_stage2_start_map
* ...
*/
static inline uint8_t engine_kpa8(void)
{
	return engine_kpa8_of( engine.kpa250 );
}

#ifndef IGN_BALANCE
#ifdef ALS
// 0.25 crankdeg resolution: 64 degree is fine
#define IGN_BALANCE 256
#else
#define IGN_BALANCE 0
#endif
#endif

#define config_iac_ignsoftidle() (config.iac_conf & _BV(7))
#define config_iac_pid_asymetric() (config.iac_conf & _BV(6))

// freeing config.baro:
#define config_baro() (100)

#define config_egt1_offs() (0)

// Reusing config variables
#ifdef ALS
#define RV_als_enrich			config.als_enrich	// als,fuel enrichment [0.4%]
#define RV_als_retard_mintps	config.als_retard_mintps	// als,min TPS for delay spark
#define RV_als_retard_maxtps	config.als_retard_maxtps	// als,max TPS for delay spark
#define RV_als_cut_mintps		config.als_cut_mintps 	// als,min TPS for cut spark
#define RV_als_cut_maxtps		config.als_cut_maxtps	// als,max TPS for cut spark
#define RV_als_react_tps		config.als_react_tps	// als,reactivate TPS
#define RV_als_tks_output		config.als_tks_channel	// als,throttle kicker output channel
#define RV_als_max_cut			config.als_max_cut // als,maximum cut strength [0..16]

#define CONF_ALS_RETARD_RPM1 config.als_retard_rpm1
#define CONF_ALS_RETARD_RPM2 config.als_retard_rpm2
#define CONF_ALS_IGNCUT_RPM1 config.als_igncut_rpm1
#define CONF_ALS_IGNCUT_RPM2 config.als_igncut_rpm2

#endif

#ifdef SOFTIDLE
#ifndef SOFTRPM
#error "Define SOFTRPM to use softrpm based idle control!"
#endif
// Idle control by variable spark cut
// config.iac_conf bit 7 to enable
#define RV_softrpm_idle_tps		actual_tps_thres // idle,sparkcut,TPS treshold
#define RV_softrpm_idle_rpm		iac.target_rpm // idle RPM target [10 RPM]
// #define RV_softrpm_idle_range	50 // * 4 RPM = 200 RPM idle,sparkcut,RPM control range
#endif // SOFTIDLE

#ifdef SOFTREDLINE
// Soft rpm limiter (redline)
#ifndef SOFTRPM
#error "Define SOFTRPM to use softrpm based redline!"
#endif
#define RV_softrpm_rpm			config.softrpm_rpm  // redline,sparkcut,RPM treshold [100 RPM]
#define RV_softrpm_range		config.softrpm_range // redline,sparkcut,RPM range [4 RPM]
#endif // SOFTREDLINE

#ifdef SOFTLAUNCH
// Launch control by spark cut
#ifndef SOFTRPM
#error "Define SOFTRPM to use launch control!"
#endif
#define RV_launch_rpm			config.launch_rpm	// launch,sparkcut,RPM treshold [100 RPM]
#ifndef SOFTREDLINE
// Using redline's softrpm range, but can be used together
#define RV_softrpm_range		config.softrpm_range	// launch,sparkcut,RPM range [4 RPM]
#endif	// SOFTREDLINE
#endif  // SOFTLAUNCH


/***********************************************************************/
// config.shiftcut_conf_input
// 7..4 are flags
// 0..3 is mcp3208 ch selector
#define CUT_SPARK_REV 6
#define CUT_SPARK_SHIFT 5
#define CUT_FUEL_SHIFT 4

//tpsdot_kpadot_conf
#define tps_or_kpa		0  // 0:tpsdot, 1:mapdot

/***********************************************************************/
// alternate
#define BATCH_FIRE_CRANKING 4 // 1:1:batch fire when cranking
//#define ADD_FUEL_CRANKING 5 // 1:1:additional fuel when cranking

/***********************************************************************/
//config11
//bit 7-4 =  number of cylinders-1
#define NCYL_1 (((config.config11 & 0xF0) >> 4))  // ncyl -1
// see CHRYSLERHEMI before changing
#define NCYL (((config.config11 & 0xF0) >> 4) + 1)  // ncyl
//note that #define NCYL ((config.config11 >> 4) + 1) or even ((config.config11 & 0xF) + 1)  would NOT save any flash
#define engine_cylinders NCYL // Use this instead of NCYL. It might be cached later.
#define engine_cylinders4 (NCYL*4)

// bit3 used to be named "port injection" vs. "TBI" injection.s
// ALS takes priority over launch when both enabled (launch disabled when ALS is activated)
#define ALS_OVER_LAUNCH ( config.config11 & 0x08 )

// bit2=1 used to be named "2stroke" but was never used that way...
// don't forget to also set config13 bit 1 (O2_WB_SENSOR) =1 !!!
#define WBO2_BOTH_CHANNELS ( config.config11 & 0x04 )
#define USE_INJECTOR_DEADTIME_SIMPLE ( config.config11 & 0x02 )
#define fix_retard_strategy() ( config.config11 & 0x01 )
// same flashsize as: (NCYL << 2) or
// #define engine_cylinders4 ((((config.config11 & 0xF0) >> 4) << 2) + 4)
// and smaller than:
// #define engine_cylinders4 (((config.config11 & 0xF0) >> 2) + 4)
// apparently >>4 is much better than other numbers (using swap)

/***********************************************************************/

//config12
#define ALTERNATIVE_RT_DATA	0	// 1:knock data 0:adc data
#define sd_bytestuffed_bit 1 // sd log is in bytestuffed format
#define ALPHAN_BARO			2	// 1:MAP-hose after throttle 0: MAP-hose before throttle
#define iac_integral_frozen_bit 3 //iac integral frozen bit, if bit is set and engine is in idle and engine_kpa8() < config.iac_integral_lowmap then iac.integral is frozen (doesn't change) But when isn't set, and ngine is in idle and ngine_kpa8() < config.iac_integral_lowmap then iac.integral is set to 0
#define iac_integral_moving_bit 4 //iac integral moving bit, when is set and cars is moving (car speed from speed sensor>0) and car is in idle then iac.integral is set to 0
#define aim_speedsensor_divider_bit 5 //Speedsensor Aim value divided by 10
#define dirac_elimination_sec_bit 6  // filter transient noise to secondary trigger
#define dirac_elimination_prim_bit 7  // filter transient noise to primary trigger
/***********************************************************************/

//config13
#define ODDFIRE               0 // 0:normal, 1:odd-fire
#define O2_WB_SENSOR          1 // 0:narrowband, 1:diy-wb
#define ALPHAN     2 // 0:use the MAP or tps map blend for fuel calcs 1:do not use any multipliers
#define BARO_CORRECTION       3 // 0:off, 1:on
#define MAT_RETARD            4 // 0:on, 1:off
#define IGNLAMBDA_TPSKPA	5 // 0: ign, lambda looked up by kpa with tps lookup under nonve_kpalookup_minkpa 1: ign, lambda looked up by tpskpa
#define CLT_CURVE_17_POINT		6	// 1:use config curve, 0: use flash curve
#define MAT_CURVE_17_POINT		7	// 1:use config curve, 0: use flash curve

// Alphan, but prefer kpa (not kpa_tps blend) for ignadv lambda. Use this when you have MAP sensor
// In other words ignadv and lambda looked up by kpa
#define ALPHAN_PREFER_KPA ((config.config13 & _BV(IGNLAMBDA_TPSKPA)) == 0 )

#define WBO2_SENSOR_CONFIGURED (config.config13 & _BV(O2_WB_SENSOR))
/***********************************************************************/
// master_debug
#define mat2map        0    // reroute the mat sensor value to the map sensor value

/***********************************************************************/
// ve_conf
#define ve_learn_on    0   // enable the VE learning
#define ve_learn_sim   1   // only simulate VE learning
#define ve_learn_2x2   2   // modify all four elements

/***********************************************************************/
// lcd_c0
#define lcd_busypolled	0	// if 1, the busy flag is polled
#define lcd_easygoto	1 	// if 0, we force goto after each data
#define pwmbrightness	3	// if 1 then the lcd uses PWM to control brightness
#define swbrightness	4	// if 1 then the lcd allows software brightness commands (Noritake VFD)
/***********************************************************************/
// trigger1 primary_trigger
#define trigger_edge   0       // if 1, then rising edge
#define ptrig_coil     1       // if 1, then use coil to trigger instead of toothwheel
#define trigger_filter 2       // if 1, then filtering is enabled
#define trigger_multitooth_hairy 3	// if 1, then noise starts a new resync sequence
#define trigger_multitooth_Nminus1 4	// if 1, then N-1 wheel (0: N-2). For (coiltype + fiatstilo/subaru + advanced missingtooth filter) it means "syncafterLONGgap", useful for Maserati
#define trigger_fiatstilo 5	// if 1, then Fiat Stilo style 60/90/30 degree trigger assumed
#define trigger_auditrigger 6	// if 1, then Audi 135+ tooth trigger style assumed
#define trigger_diveby2 7	// 0=divideby3, 1=divideby2

// pass config.tooth_wheel as argument
#define is_elise_trigger(ctw) (ctw == 32)
// NCYL==8 inside is_elise_trigger() means Chrysler hemi http://www.dynotech.se/bilder/hemi_trigger.jpg :
#define CHRYSLERHEMI   ((config.config11 & 0xF0)==0x70)

// subaru 36-2-2-2 (derived from c012 or c024? Certainly not c004) ONLY when FIATSTILO_SUBARU is also set !
#define is_subaru36222_trigger(ctw) (ctw >= 12)
// in FIATSTILO_SUBARU (but NOT in LANCIA_COSWORTH !) another_trigger_tooth refers to CAM, so normally use: catt = CONF_another_trigger_tooth
#define CONF_another_trigger_tooth ((FIATSTILO_SUBARU && ! is_subaru36222_trigger(config.tooth_wheel)) ? 1 : config.another_trigger_tooth)

#define THIRD_TRIGGER_INTERRUPT_ENABLED (config_ext.vvti_cold & 0x08)

#define is_nisssan_divider() (config_ext.strig_validation & _BV(NISSAN_DIVIDER) )
#define is_nissan_trigger() (config_ext.strig_validation & _BV(NISSAN_TRIGGER) )
#define is_suzuki_trigger() (config_ext.strig_validation & _BV(SUZUKI_TRIGGER) )
#define is_extra_pulse_trigger() ( config_ext.strig_validation & _BV(EXTRA_PULSE_TRIGGER) )
#define is_cam_check_pol() (  config_ext.strig_validation & _BV(CAM_CHECK_POLARITY) )

// trigger2 (bit 0 and 2 are named SAME as for primary_trigger)
//#define trigger_edge   0       // if 1, then rising edge
#define strig_disable    1       // if 1, then secondary trigger is disabled. Only use for interrupt enable ! Otherwise examine strig_cam_sync !
//#define trigger_filter 2       // if 1, then filtering is enabled
#define strig_secondary  3       // if 1, then use for secondary trigger (as traditionally). If 0 then use sectrig input for primary trigger
#define strig_cam_sync   4       // if 1, then enable cam sync, else measure alien ignition advance
#define strig_both_edges 5       // if 1, then cam sync on both edges during cranking
#define strig_subaruspecial 6       // if 1, camwheel subaruspecial (must be enabled for subaru, it works either way for fiatstilo -)
#define audi_hall_scope_or_missing_sync_without_camsync 7 //audi hall scope enable bit if auditrigger used or missing tooth sync without cam received enable bit. Never use with toyota defines
#define strig_vvti       7       // if 1, then use toyota 1JZ VVT-i control, everywhere under toyota defines

/***********************************************************************/
// knock_conf WAS: (now changed)
// #define knock_enable   0   // if 1, then enabled
// #define knock_chan     1   // if 1, then channel 1 else channel 0

/***********************************************************************/

// ign_out definitions:
#define IGN_DISABLE		(7)
#define IGN_DUMMYDWELL	(6)
// bit5=0 for CDI
#define IGN_BEFORE		(5)
#define IGN_REFSPARK	(4)
#define IGN_MULTISPARK	(3)
#define IGN_DUAL_OUT	(1)
#define IGN_INVERTOUT	(0)
// used combinations of upper nibble (7..4) bits:
#define IGNTYPE_EDIS	(0x00)
#define IGNTYPE_DUMMY	(0x70)
// the rest 14 combinations are off, but:
#define IGNTYPE_NONE	(0xF0)

// boost_conf definitions:
// bit1: wheelspeed dependent
#define BOOST_WHEELSPEED_MASK (2)
// bit2:
#define BOOST_GEARDEPENDANT_MASK (4)
//bit3: reducing refDC (not just indirectly via boosttarget) when egt or IAT is too high:
#define BOOST_EGT_MAT_REFDC_REDUCING_MASK (8)
// boost_conf bit 1:0 define 4 modes of operation:
#define BOOSTCONF_MODEMASK (1)

#define BOOSTCONF_RELATIVEMODE_CONFIGURED ((config.boost_conf & BOOSTCONF_MODEMASK) == 0)

// boost_more definitions:
#define BOOST_TARGETOFFS (7)
// was BOOSTTARGET (from nernst2), now PGASLAMBDATARGET
#define PGASLAMBDATARGET (6)
#define BOOSTTARGET      (5)

#define BOOST_TARGET_MASK ( _BV(PGASLAMBDATARGET) | _BV(BOOSTTARGET ) ) 
#define BOOST_TARGET      ( _BV(PGASLAMBDATARGET) | _BV(BOOSTTARGET ) )     

#ifndef KPAFAC_UNIT
// 2 kPa is default
#define KPAFAC_UNIT (2)
#endif

// should be in <avr/pgmspace.h> but it's missing from older avrlibc:

#ifndef pgm_read_byte
#define pgm_read_byte(a) (PRG_RDB((unsigned char*)(a)))
#endif

#ifndef PRG_RDB
#define PRG_RDB(x) pgm_read_byte(x)
#endif

#ifndef pgm_read_word
// TODO: this assumes little-endian: make it endian-safe
#define pgm_read_word(a) ((PRG_RDB((unsigned char*)(a)) & 0x00FF) | ((PRG_RDB((unsigned char*)(a)+1))<<8))
#endif

#ifndef pgm_read_word_near
#define pgm_read_word_near(a) pgm_read_word(a)
#endif

#ifndef pgm_read_addr
#define pgm_read_addr(a) pgm_read_word(a)
#endif

// deprecated. Use the ones from above:
#ifndef PRG_RDW
#define PRG_RDW(x)  pgm_read_word(x)
#endif


#ifndef pointer_to_upper8_16
// hack, only suits little endian. TODO: avr-libc solution
#define pointer_to_upper8_16(x) ((uint8_t*)(x) + 1)
#define pointer_to_lower8_16(x) ((uint8_t*)(x))
#endif

extern const uint8_t BV_rev_arr[];

// similar to BV macro but reverse bitorder (0 => 0x80)
// extern uint8_t BV_rev_arr[] PROGMEM ; // = {0x80, 0x40, 0x20, 0x10, 8, 4, 2, 1};
#define BV_rev(a) (pgm_read_byte(&BV_rev_arr[a]))
#define BV_rev_l(a) (pgm_read_byte(&BV_rev_arr[a&7]))

// macro for putting the code/variables into defined sections
#define SECTION(_sect) __attribute__((section(_sect)))

#if IGNCHANNELS>8
#define INDIVIDUAL_POWER_SIZE IGNCHANNELS
#else
#define INDIVIDUAL_POWER_SIZE HWMAP_SIZE
#endif
EXTERN uint8_t individual_power[INDIVIDUAL_POWER_SIZE];

/**
 * Steps down, except after 0: than step back to ignchmax
 * @param ic
 * @return
 */
static inline uint8_t next_igncount(uint8_t i)
{
	uint8_t ignchmax;
	if(i) return i-1;
	// i was 0, underflow:
	ignchmax = config.ignchmax;
	if(ignchmax==4 && FIATSTILO_SUBARU && VIPER) return 9; // dire consequences ?
	
	return ignchmax;
}
/** idx = new cyl value, that is next_igncount( get_ignumsc() ) */
static inline void save_ignumsc(uint8_t idx)
{
        set_ignumsc( idx );
	if( engine.status_ext & _BV(individual_allowed) ){
		individual_power[idx]= engine.individual_power;
	}
}
/** returns the OLD value */
static inline uint8_t step_ignumdc(void)
{
        uint8_t idx = get_ignumdc();
	set_ignumdc( next_igncount(idx) );
	return idx; // return the OLD value
}

static inline uint8_t get_ignum(void)
{
	uint8_t idx = get_ignumsc();
	idx <<= 4;
	idx |= get_ignumdc();
	return idx;
}
#if IGNCHANNELS>8
// using 1 byte for each bit (later maybe timestamp or activated output ?)
static inline void set_cyl_dwelling(uint8_t cyl ){
	engine.ignum_dwellarr[cyl] = 1;
}

static inline void reset_cyl_dwelling(uint8_t cyl ){
	engine.ignum_dwellarr[cyl] = 0;
}

static inline uint8_t is_cyl_dwelling(uint8_t cyl ){
	return engine.ignum_dwellarr[cyl];
}
static inline void init_cyl_dwelling(uint8_t cyl ){
  ;
}
#else
static inline void set_cyl_dwelling(uint8_t cyl )
{
	engine.ignum_dwell |= BV_rev( cyl );
}

static inline void reset_cyl_dwelling(uint8_t cyl )
{
	engine.ignum_dwell &= ~ BV_rev( cyl );
}

static inline uint8_t is_cyl_dwelling(uint8_t cyl )
{
	return engine.ignum_dwell & BV_rev( cyl );
}

static inline void init_cyl_dwelling(uint8_t cyl )
{
	engine.ignum_dwell = cyl;
}
#endif

static inline uint8_t get_injnum(void){
        return  engine.injnum;
}

static inline void reset_injnum(uint8_t newval){
	engine.injnum = newval;
}

static inline uint8_t next_injnum( uint8_t inj)
{
	if( inj == 0 ) return PRIMARY_INJ_LAST;
	return inj-1;
}

static inline void save_injnum(uint8_t inj)
{
    if ( inj > PRIMARY_INJ_LAST ) inj = 0;// prevent bad value
    engine.injnum = inj;
}

/**
 returns the number of teeth, together with the missing.
 multiplied by 2 if camsync is used.
 15 - 1 with camsync : 30 (config.tooth_wheel == 14)
 60 - 2 with camsync: 120 (config.tooth_wheel == 58)
 60 - 2 without camsync: 60 (config.tooth_wheel == 58)
 36 - 1 without camsync: 36 (config.tooth_wheel == 35)
 When ptrig_coil then it is understood that we have a reset on the secondary
 trigger and config.tooth_wheel is understood to be the number of teeth
 between resets (normally taken from the cam) in this case we don't need *2
 cnfig.tooth_wheel MUST be set for coiltype as well, see
 http://www.vems.hu/wiki/index.php?page=InputTrigger%2FMultiToothNoneMissing

*/
static inline uint8_t get_total_tooth(void)
{
	uint8_t cpt, t8 = config.tooth_wheel;
	cpt = config.primary_trigger;
	if (!is_extra_pulse_trigger()){//extrapulse trigger needs just *2 if cam sync use 
        if( cpt & _BV(ptrig_coil) ) {
            // no <<= 1 (even with camsync!) So for simple trigger configure 4 for 4cyl, not 2!
            return t8;
        }

        if ( is_elise_trigger(t8)){
            t8+=4;
            return t8;
        }
        if( cpt & _BV(trigger_multitooth_Nminus1) ) {
            // N-1:
            t8 ++;
        } else { // N-2
            t8 += 2;
        }
	}

	if ( config.secondary_trigger & _BV(strig_cam_sync)
		|| config.tooth_wheel <= 3 ){ // camsyncpretender for the 4-1 wheel
		// camsync configured, so *2: t8 <<= 1; t8 += t8;
		t8 <<= 1;
	}

	return t8;
}

#include "eventqueue.h"
/** should engine_off_delay==0 mean "never" ? or "almost immediately" ? (previously it was a bad setting) */
static inline uint8_t engine_off_delay_has_passed( void )
{
	uint8_t n8;
	uint16_t now_upper = etime.now_upper;
	if(now_upper <= config.pump_on_mintime) return 0; // initial few seconds
	now_upper -= (etime.ptrig_capturedtime_prev>>16); // time since trigger pulse
	if(now_upper >> 8) return 1; // >= 256

	n8 = now_upper & 0xFF;
	if( n8 >= 2 && n8 > config.engine_off_delay) return 1; // 0 or 1 is too low.
	return 0; // not yet
}

//prohibit_status:
#define INJ_PROHIBIT 0
#define IGN_PROHIBIT 1
#define STEPPER_PROHIBIT 2

#define o2_closedloop   3  // enabled when o2 sensor is active

#define PROHIBIT_SECPOS 7
#define PROHIBIT_MASK (_BV(INJ_PROHIBIT) | _BV(IGN_PROHIBIT) | _BV(STEPPER_PROHIBIT))

static inline uint8_t isInjProhibited(void)
{
    return engine.prohibit_status & _BV(INJ_PROHIBIT);
}

static inline uint8_t isIgnProhibited(void)
{
    return engine.prohibit_status & _BV(IGN_PROHIBIT);
}

static inline uint8_t isStepperProhibited(void)
{
    return engine.prohibit_status & _BV(STEPPER_PROHIBIT);
}

static inline uint8_t isO2_closedloop(void)
{
    return engine.prohibit_status & _BV(o2_closedloop);
}

#endif	// GLOBAL_H
